// Flame Screensaver for Windows by John Sullivan
//
// Screen hack code taken (with modifications) from xscreensaver package
// by Jamie Zawinski. Original Copyright and License terms from xscreensaver
// package follow:

/* xscreensaver, Copyright (c) 1993, 1995, 1996, 1998
 *  Jamie Zawinski <jwz@jwz.org>
 *
 * Permission to use, copy, modify, distribute, and sell this software and its
 * documentation for any purpose is hereby granted without fee, provided that
 * the above copyright notice appear in all copies and that both that
 * copyright notice and this permission notice appear in supporting
 * documentation.  No representations are made about the suitability of this
 * software for any purpose.  It is provided "as is" without express or 
 * implied warranty.
 */

/* This file was ported from xlock for use in xscreensaver (and standalone)
 * by jwz on 18-Oct-93.  (And again, 11-May-97.)  Original copyright reads:
 *
 *   static char sccsid[] = "@(#)flame.c 1.4 91/09/27 XLOCK";
 *
 * flame.c - recursive fractal cosmic flames.
 *
 * Copyright (c) 1991 by Patrick J. Naughton.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose and without fee is hereby granted,
 * provided that the above copyright notice appear in all copies and that
 * both that copyright notice and this permission notice appear in
 * supporting documentation.
 *
 * This file is provided AS IS with no warranties of any kind.  The author
 * shall have no liability with respect to the infringement of copyrights,
 * trade secrets or any patents by this file or any part thereof.  In no
 * event will the author be liable for any lost revenue or profits or
 * other special, indirect and consequential damages.
 *
 * Comments and additions should be sent to the author:
 *
 *             naughton@eng.sun.com
 *
 *             Patrick J. Naughton
 *             MS 21-14
 *             Sun Laboritories, Inc.
 *             2550 Garcia Ave
 *             Mountain View, CA  94043
 *
 * Revision History:
 * 01-Jun-95: This should look more like the original with some updates by
 *            Scott Draves.
 * 27-Jun-91: vary number of functions used.
 * 24-Jun-91: fixed portability problem with integer mod (%).
 * 06-Jun-91: Written. (received from Scott Draves, spot@cs.cmu.edu).
 */

#include "stdafx.h"
#include "hack.h"

#include <math.h>

static int flame_inited = 0;
static HWND hwndFirst = NULL;
static flame_struct fs;

unsigned int random();
void make_smooth_colormap (LOGPALETTE *pal, int *ncolorsP);


static short
halfrandom (int mv)
{
  static short lasthalf = 0;
  unsigned long r;

  if (lasthalf)
    {
      r = lasthalf;
      lasthalf = 0;
    }
  else
    {
      r = random ();
      lasthalf = (short)(r >> 16);
    }
  return (short)(r % mv);
}

flame_struct*
init_flame (HWND hwnd)
{
    flame_struct *p = new flame_struct;
	memset(p, 0, sizeof(*p));

    flame_struct *q = &fs;

    RECT rc;
    ::GetClientRect(hwnd, &rc);
    p->width = rc.right;
    p->height = rc.bottom;

	p->max_points = get_integer_resource ("iterations", "Integer");
	if (p->max_points <= 0) p->max_points = 100;

	p->max_levels = p->max_points;

	p->max_total = get_integer_resource ("points", "Integer");
	if (p->max_total <= 0) p->max_total = 10000;

	p->delay = get_integer_resource ("delay", "Integer");
	if (p->delay < 0) p->delay = 0;
	p->delay2 = get_integer_resource ("delay2", "Integer");
	if (p->delay2 < 0) p->delay2 = 0;

//  if (p->delay<50) p->delay = 50; // minimum

	p->variation = random() % MAXKINDS;

    if (!flame_inited) {
        memset(&fs, 0, sizeof(fs));

	    q->ncolors = get_integer_resource ("colors", "Integer");
	    if (q->ncolors <= 0) q->ncolors = 128;

	    q->colors = (LOGPALETTE*) malloc ((q->ncolors+1) * sizeof (PALETTEENTRY) + sizeof(LOGPALETTE));
	    memset(q->colors, 0, (q->ncolors+1) * sizeof (PALETTEENTRY) + sizeof(LOGPALETTE));
	    q->colors->palVersion = 0x300;
	    q->colors->palNumEntries = q->ncolors+1;
	    make_smooth_colormap (q->colors, &q->ncolors);
	    q->flamepal = CreatePalette(q->colors);

	    q->pixcol = 0;

        p->first = true;
        hwndFirst = hwnd;
        flame_inited = 1;
    }

    return p;
}

static int
recurse (flame_struct *p, flame_struct *q, double x, double y, int l, HWND hwnd, HDC hdc)
{
    int xp, yp, i;
    double nx, ny;

    if (l == p->max_levels) {
        p->total_points++;
        if (p->total_points > p->max_total) /* how long each fractal runs */
            return 0;

        if (x > -1.0 && x < 1.0 && y > -1.0 && y < 1.0)
        {
            xp = p->points[p->num_points].x = (int) ((p->width / 2) * (x + 1.0));
            yp = p->points[p->num_points].y = (int) ((p->height / 2) * (y + 1.0));
            p->num_points++;
            if (p->num_points >= POINT_BUFFER_SIZE)
            {
                for ( int ptl=0 ; ptl<p->num_points ; ptl++ ) {
                    SetPixelV(hdc, p->points[ptl].x, p->points[ptl].y, q->cur_col);
                }
                p->num_points = 0;
                /* if (delay) usleep (delay); */
                /* XSync (dpy, False); */
            }
        }
    }
    else
    {
        for (i = 0; i < p->snum; i++)
        {

        /* Scale back when values get very large. Spot sez:
         "I think this happens on HPUX.  I think it's non-IEEE
         to generate an exception instead of a silent NaN."
        */
            if ((fabs(x) > 1.0E5) || (fabs(y) > 1.0E5))
                x = x / y;

            nx = p->f[0][0][i] * x + p->f[0][1][i] * y + p->f[0][2][i];
            ny = p->f[1][0][i] * x + p->f[1][1][i] * y + p->f[1][2][i];
            if (i < p->anum)
            {
                switch (p->variation)
                {
                case 0: /* sinusoidal */
                    nx = sin(nx);
                    ny = sin(ny);
                    break;
                case 1: /* complex */
                    {
                        double r2 = nx * nx + ny * ny + 1e-6;
                        nx = nx / r2;
                        ny = ny / r2;
                    }
                    break;
                case 2: /* bent */
                    if (nx < 0.0)
                        nx = nx * 2.0;
                    if (ny < 0.0)
                        ny = ny / 2.0;
                    break;
                case 3: /* swirl */
                    {
                        double r = (nx * nx + ny * ny); /* times k here is fun */
                        double c1 = sin(r);
                        double c2 = cos(r);
                        double t = nx;

                        if (nx > 1e4 || nx < -1e4 || ny > 1e4 || ny < -1e4)
                            ny = 1e4;
                        else
                            ny = c2 * t + c1 * ny;
                        nx = c1 * nx - c2 * ny;
                    }
                    break;
                case 4: /* horseshoe */
                    {
                        double r, c1, c2, t;

                        /* Avoid atan2: DOMAIN error message */
                        if (nx == 0.0 && ny == 0.0)
                            r = 0.0;
                        else
                            r = atan2(nx, ny);      /* times k here is fun */
                        c1 = sin(r);
                        c2 = cos(r);
                        t = nx;

                        nx = c1 * nx - c2 * ny;
                        ny = c2 * t + c1 * ny;
                    }
                    break;
                case 5: /* drape */
                    {
                        double t;

                        /* Avoid atan2: DOMAIN error message */
                        if (nx == 0.0 && ny == 0.0)
                            t = 0.0;
                        else
                            t = atan2(nx, ny) / M_PI;

                        if (nx > 1e4 || nx < -1e4 || ny > 1e4 || ny < -1e4)
                            ny = 1e4;
                        else
                            ny = sqrt(nx * nx + ny * ny) - 1.0;
                        nx = t;
                    }
                    break;
                case 6: /* broken */
                    if (nx > 1.0)
                        nx = nx - 1.0;
                    if (nx < -1.0)
                        nx = nx + 1.0;
                    if (ny > 1.0)
                        ny = ny - 1.0;
                    if (ny < -1.0)
                        ny = ny + 1.0;
                  break;
                case 7: /* spherical */
                    {
                        double r = 0.5 + sqrt(nx * nx + ny * ny + 1e-6);

                        nx = nx / r;
                        ny = ny / r;
                    }
                    break;
                case 8: /*  */
                    nx = atan(nx) / M_PI_2;
                    ny = atan(ny) / M_PI_2;
                    break;
/* #if 0 */  /* core dumps on some machines, why not all? */
                case 9: /* complex sine */
                    {
                        double u = nx;
                        double v = ny;
                        double ev = exp(v);
                        double emv = exp(-v);

                        nx = (ev + emv) * sin(u) / 2.0;
                        ny = (ev - emv) * cos(u) / 2.0;
                    }
                    break;
                case 10:    /* polynomial */
                    if (nx < 0)
                        nx = -nx * nx;
                    else
                        nx = nx * nx;
                    if (ny < 0)
                        ny = -ny * ny;
                    else
                        ny = ny * ny;
                    break;
/* #endif */
                default:
                    nx = sin(nx);
                    ny = sin(ny);
                }
            }
            if (!recurse (p, q, nx, ny, l + 1, hwnd, hdc))
                return 0;
        }
    }
    return 1;
}


bool
flame (flame_struct *p, HWND hwnd, HDC hdc)
{
    int i, j, k;
    static int alt = 0;

    flame_struct *q = &fs;

    SelectPalette(hdc, q->flamepal, TRUE);
    SelectPalette(hdc, q->flamepal, FALSE);
    RealizePalette(hdc);

    if (!(p->cur_level++ % p->max_levels))
    {
//      if (p->delay2) usleep (p->delay2);
//      if (p->delay2) Sleep (p->delay2);
        {
            RECT rc;
            rc.top = rc.left = 0;
            rc.right = p->width;
            rc.bottom = p->height;
            FillRect(hdc, &rc, (HBRUSH)GetStockObject(BLACK_BRUSH));
        }
        alt = !alt;

        p->variation = random() % MAXKINDS;
    }
    else
    {
        if (q->ncolors > 2 && hwnd==hwndFirst)
        {
            q->cur_col = *(COLORREF*)(q->colors->palPalEntry+q->pixcol);
            if (--q->pixcol < 0)
                q->pixcol = q->ncolors - 1;
        }
    }

    /* number of functions */
    p->snum = 2 + (p->cur_level % (MAXLEV - 1));

    /* how many of them are of alternate form */
    if (alt)
        p->anum = 0;
    else
        p->anum = halfrandom (p->snum) + 2;

    /* 6 coefs per function */
    for (k = 0; k < p->snum; k++)
    {
        for (i = 0; i < 2; i++)
            for (j = 0; j < 3; j++)
                p->f[i][j][k] = ((double) (random() & 1023) / 512.0 - 1.0);
    }
    p->num_points = 0;
    p->total_points = 0;
    (void) recurse (p, q, 0.0, 0.0, 0, hwnd, hdc);

    for ( int ptl=0 ; ptl<p->num_points ; ptl++ ) {
        SetPixelV(hdc, p->points[ptl].x, p->points[ptl].y, q->cur_col);
    }

    return !((p->cur_level+1) % p->max_levels);
}

struct {
    char *name;
    int defval;
    int val;
    bool loaded;
    bool modified;
} prefs[] = {
    { "colors", 64, 0, false, false },
    { "iterations", 25, 0, false, false },
    { "delay", 50, 0, false, false },
    { "delay2", 2000, 0, false, false },
    { "points", 10000, 0, false, false },
    { NULL, 0, 0, false, false }
};

void load_prefs()
{
    USES_CONVERSION;

    HKEY hk=NULL;
    if (RegOpenKeyEx(HKEY_CURRENT_USER, _T("Control Panel\\Screen Saver.Flame"), 0, KEY_READ, &hk)==ERROR_SUCCESS) {
        for ( int i=0 ; prefs[i].name ; i++ ) {
            DWORD dwType=0;
            DWORD dwVal=-1;
            DWORD dwLen = sizeof(DWORD);

            if (RegQueryValueEx(hk, A2T(prefs[i].name), 0, &dwType, (LPBYTE)&dwVal, &dwLen)==ERROR_SUCCESS && dwType==REG_DWORD) {

                prefs[i].val = (int)dwVal;
                prefs[i].loaded = true;

            } else {
                prefs[i].val = prefs[i].defval;
            }
        }

        RegCloseKey(hk);
    }
}

void save_prefs()
{
    USES_CONVERSION;

    HKEY hk=NULL;

    for ( int i=0 ; prefs[i].name ; i++ ) {
        if (prefs[i].modified) {
            if (!hk) {
                if (RegCreateKeyEx(HKEY_CURRENT_USER, _T("Control Panel\\Screen Saver.Flame"), 0, _T(""), REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hk, NULL)!=ERROR_SUCCESS) {
                    return;
                }
            }

            DWORD dwVal = (DWORD)prefs[i].val;

            if (RegSetValueEx(hk, A2T(prefs[i].name), 0, REG_DWORD, (const BYTE *)&dwVal, sizeof(DWORD))==ERROR_SUCCESS) {
                prefs[i].modified = false;
            }
        }
    }

    if (hk) RegCloseKey(hk);
}

void reset_prefs()
{
    for ( int i=0 ; prefs[i].name ; i++ ) {
        prefs[i].loaded = prefs[i].modified = false;
    }
}

int get_integer_resource(char *name, char *dummy)
{
    for ( int i=0 ; prefs[i].name ; i++ ) {
        if (strcmp(prefs[i].name, name)==0) {
            if (prefs[i].loaded || prefs[i].modified) {
                return prefs[i].val;
            } else {
                return prefs[i].defval;
            }
        }
    }
    return -1;
}

void put_integer_resource(char *name, int val)
{
    for ( int i=0 ; prefs[i].name ; i++ ) {
        if (strcmp(prefs[i].name, name)==0) {
            if (prefs[i].val!=val) {
                prefs[i].val = val;
                prefs[i].modified = true;
            }
        }
    }
}

char *defaults [] = {
  ".background:	black",
  ".foreground:	white",
  "*colors:	64",
  "*iterations:	25",
  "*delay:	50000",
  "*delay2:	2000000",
  "*points:	10000",
  0
};


// ya_random

/* The following 'random' numbers are taken from CRC, 18th Edition, page 622.
   Each array element was taken from the corresponding line in the table,
   except that a[0] was from line 100. 8s and 9s in the table were simply
   skipped. The high order digit was taken mod 4.
 */
#define VectorSize 55
static unsigned int a[VectorSize] = {
 035340171546, 010401501101, 022364657325, 024130436022, 002167303062, /*  5 */
 037570375137, 037210607110, 016272055420, 023011770546, 017143426366, /* 10 */
 014753657433, 021657231332, 023553406142, 004236526362, 010365611275, /* 14 */
 007117336710, 011051276551, 002362132524, 001011540233, 012162531646, /* 20 */
 007056762337, 006631245521, 014164542224, 032633236305, 023342700176, /* 25 */
 002433062234, 015257225043, 026762051606, 000742573230, 005366042132, /* 30 */
 012126416411, 000520471171, 000725646277, 020116577576, 025765742604, /* 35 */
 007633473735, 015674255275, 017555634041, 006503154145, 021576344247, /* 40 */
 014577627653, 002707523333, 034146376720, 030060227734, 013765414060, /* 45 */
 036072251540, 007255221037, 024364674123, 006200353166, 010126373326, /* 50 */
 015664104320, 016401041535, 016215305520, 033115351014, 017411670323  /* 55 */
};

static int i1, i2;

unsigned int random (void)
{
  register int ret = a[i1] + a[i2];
  a[i1] = ret;
  if (++i1 >= VectorSize) i1 = 0;
  if (++i2 >= VectorSize) i2 = 0;
  return ret;
}

void ya_rand_init(unsigned int seed)
{
  int i;

  a[0] += seed;
  for (i = 1; i < VectorSize; i++)
    {
      seed = a[i-1]*1001 + seed*999;
      a[i] += seed;
    }

  i1 = a[0] % VectorSize;
  i2 = (i1 + 024) % VectorSize;
}

static double _frand_tmp_;
# define frand(f)                           \
  (_frand_tmp_ = (((double) random()) /                 \
          (((double) ((unsigned int)~0)) / ((double) (f)))),    \
   _frand_tmp_ < 0 ? (-_frand_tmp_) : _frand_tmp_)


// colormap stuff

void
hsv_to_rgb (int h, double s, double v,
        unsigned char *r, unsigned char *g, unsigned char *b)
{
  double H, S, V, R, G, B;
  double p1, p2, p3;
  double f;
  int i;

  if (s < 0) s = 0;
  if (v < 0) v = 0;
  if (s > 1) s = 1;
  if (v > 1) v = 1;

  S = s; V = v;
  H = (h % 360) / 60.0;
  i = (int)H;
  f = H - i;
  p1 = V * (1 - S);
  p2 = V * (1 - (S * f));
  p3 = V * (1 - (S * (1 - f)));
  if      (i == 0) { R = V;  G = p3; B = p1; }
  else if (i == 1) { R = p2; G = V;  B = p1; }
  else if (i == 2) { R = p1; G = V;  B = p3; }
  else if (i == 3) { R = p1; G = p2; B = V;  }
  else if (i == 4) { R = p3; G = p1; B = V;  }
  else         { R = V;  G = p1; B = p2; }
  *r = (BYTE)(R * 255.0);
  *g = (BYTE)(G * 255.0);
  *b = (BYTE)(B * 255.0);
}


void
make_color_ramp (int h1, double s1, double v1,   /* 0-360, 0-1.0, 0-1.0 */
         int h2, double s2, double v2,   /* 0-360, 0-1.0, 0-1.0 */
         LOGPALETTE *pal, int *ncolorsP, bool closed_p)
{
  int i;
  int total_ncolors = *ncolorsP;
  int ncolors, wanted;
  double dh, ds, dv;        /* deltas */

  wanted = total_ncolors;
  if (closed_p)
    wanted = (wanted / 2) + 1;

  ncolors = total_ncolors;

  memset (pal->palPalEntry, 0, (*ncolorsP) * sizeof(PALETTEENTRY));

  if (closed_p)
    ncolors = (ncolors / 2) + 1;

  /* Note: unlike other routines in this module, this function assumes that
     if h1 and h2 are more than 180 degrees apart, then the desired direction
     is always from h1 to h2 (rather than the shorter path.)  make_uniform
     depends on this.
   */
  dh = ((double)h2 - (double)h1) / ncolors;
  ds = (s2 - s1) / ncolors;
  dv = (v2 - v1) / ncolors;

  for (i = 0; i < ncolors; i++)
    {
      pal->palPalEntry[i].peFlags = 0;
      hsv_to_rgb ((int) (h1 + (i*dh)), (s1 + (i*ds)), (v1 + (i*dv)),
          &pal->palPalEntry[i].peRed, &pal->palPalEntry[i].peGreen, &pal->palPalEntry[i].peBlue);
    }

  if (closed_p)
    for (i = ncolors; i < *ncolorsP; i++)
      pal->palPalEntry[i] = pal->palPalEntry[(*ncolorsP)-i];

    return;

}


#define MAXPOINTS 50    /* yeah, so I'm lazy */

static void
make_color_path (int npoints, int *h, double *s, double *v,
         LOGPALETTE *pal, int *ncolorsP)
{
  int i, j, k;
  int total_ncolors = *ncolorsP;

  int ncolors[MAXPOINTS];  /* number of pixels per edge */
  double dh[MAXPOINTS];    /* distance between pixels, per edge (0 - 360.0) */
  double ds[MAXPOINTS];    /* distance between pixels, per edge (0 - 1.0) */
  double dv[MAXPOINTS];    /* distance between pixels, per edge (0 - 1.0) */

  if (npoints == 0)
    {
      *ncolorsP = 0;
      return;
    }
  else if (npoints == 2)    /* using make_color_ramp() will be faster */
    {
      make_color_ramp (h[0], s[0], v[0], h[1], s[1], v[1],
               pal, ncolorsP, true);
      return;
    }
  else if (npoints >= MAXPOINTS)
    {
      npoints = MAXPOINTS-1;
    }

 AGAIN:

  {
    double DH[MAXPOINTS];   /* Distance between H values in the shortest
                   direction around the circle, that is, the
                   distance between 10 and 350 is 20.
                   (Range is 0 - 360.0.)
                */
    double edge[MAXPOINTS]; /* lengths of edges in unit HSV space. */
    double ratio[MAXPOINTS];    /* proportions of the edges (total 1.0) */
    double circum = 0;
    double one_point_oh = 0;    /* (debug) */

    for (i = 0; i < npoints; i++)
      {
    int j = (i+1) % npoints;
    double d = ((double) (h[i] - h[j])) / 360;
    if (d < 0) d = -d;
    if (d > 0.5) d = 0.5 - (d - 0.5);
    DH[i] = d;
      }

    for (i = 0; i < npoints; i++)
      {
    int j = (i+1) % npoints;
    edge[i] = sqrt((DH[i] * DH[j]) +
               ((s[j] - s[i]) * (s[j] - s[i])) +
               ((v[j] - v[i]) * (v[j] - v[i])));
    circum += edge[i];
      }

    if (circum < 0.0001)
      goto FAIL;

    for (i = 0; i < npoints; i++)
      {
    ratio[i] = edge[i] / circum;
    one_point_oh += ratio[i];
      }

    if (one_point_oh < 0.99999 || one_point_oh > 1.00001)
      abort();

    /* space the colors evenly along the circumference -- that means that the
       number of pixels on a edge is proportional to the length of that edge
       (relative to the lengths of the other edges.)
     */
    for (i = 0; i < npoints; i++)
      ncolors[i] = (int)(total_ncolors * ratio[i]);

    for (i = 0; i < npoints; i++)
      {
    int j = (i+1) % npoints;

    if (ncolors[i] > 0)
      {
        dh[i] = 360 * (DH[i] / ncolors[i]);
        ds[i] = (s[j] - s[i]) / ncolors[i];
        dv[i] = (v[j] - v[i]) / ncolors[i];
      }
      }
  }

  memset (pal->palPalEntry, 0, (*ncolorsP) * sizeof(PALETTEENTRY));

  k = 0;
  for (i = 0; i < npoints; i++)
    {
      int distance, direction;
      distance = h[(i+1) % npoints] - h[i];
      direction = (distance >= 0 ? -1 : 1);

      if (distance > 180)
    distance = 180 - (distance - 180);
      else if (distance < -180)
    distance = -(180 - ((-distance) - 180));
      else
    direction = -direction;

      for (j = 0; j < ncolors[i]; j++, k++)
    {
      double hh = (h[i] + (j * dh[i] * direction));
      if (hh < 0) hh += 360;
      else if (hh > 360) hh -= 0;
      pal->palPalEntry[k].peFlags = 0;
      hsv_to_rgb ((int)
              hh,
              (s[i] + (j * ds[i])),
              (v[i] + (j * dv[i])),
              &pal->palPalEntry[k].peRed, &pal->palPalEntry[k].peGreen, &pal->palPalEntry[k].peBlue);
    }
    }

  /* Floating-point round-off can make us decide to use fewer colors. */
  if (k < *ncolorsP)
    {
      *ncolorsP = k;
      if (k <= 0)
    return;
    }


  return;

 FAIL:
  /* we weren't able to allocate all the colors we wanted;
     decrease the requested number and try again.
   */
  total_ncolors = (total_ncolors > 170 ? total_ncolors - 20 :
           total_ncolors > 100 ? total_ncolors - 10 :
           total_ncolors >  75 ? total_ncolors -  5 :
           total_ncolors >  25 ? total_ncolors -  3 :
           total_ncolors >  10 ? total_ncolors -  2 :
           total_ncolors >   2 ? total_ncolors -  1 :
           0);
  *ncolorsP = total_ncolors;
  if (total_ncolors > 0)
    goto AGAIN;
}

void
make_smooth_colormap (LOGPALETTE *pal, int *ncolorsP)
{
  int npoints;
  int ncolors = *ncolorsP;
  int i;
  int h[MAXPOINTS];
  double s[MAXPOINTS];
  double v[MAXPOINTS];
  double total_s = 0;
  double total_v = 0;

  if (*ncolorsP <= 0) return;

  {
    int n = random() % 20;
    if      (n <= 5)  npoints = 2;  /* 30% of the time */
    else if (n <= 15) npoints = 3;  /* 50% of the time */
    else if (n <= 18) npoints = 4;  /* 15% of the time */
    else             npoints = 5;   /*  5% of the time */
  }

 REPICK_ALL_COLORS:
  for (i = 0; i < npoints; i++)
    {
    REPICK_THIS_COLOR:
      h[i] = random() % 360;
      s[i] = frand(1.0);
      v[i] = frand(0.8) + 0.2;

      /* Make sure that no two adjascent colors are *too* close together.
     If they are, try again.
       */
      if (i > 0)
    {
      int j = (i+1 == npoints) ? 0 : (i-1);
      double hi = ((double) h[i]) / 360;
      double hj = ((double) h[j]) / 360;
      double dh = hj - hi;
      double distance;
      if (dh < 0) dh = -dh;
      if (dh > 0.5) dh = 0.5 - (dh - 0.5);
      distance = sqrt ((dh * dh) +
               ((s[j] - s[i]) * (s[j] - s[i])) +
               ((v[j] - v[i]) * (v[j] - v[i])));
      if (distance < 0.2)
        goto REPICK_THIS_COLOR;
    }
      total_s += s[i];
      total_v += v[i];
    }

  /* If the average saturation or intensity are too low, repick the colors,
     so that we don't end up with a black-and-white or too-dark map.
   */
  if (total_s / npoints < 0.2)
    goto REPICK_ALL_COLORS;
  if (total_v / npoints < 0.3)
    goto REPICK_ALL_COLORS;

  make_color_path (npoints, h, s, v, pal, &ncolors);

  *ncolorsP = ncolors;
}

